from .exceptions.errors import NotAllowed
from .ext.selectel.selectel import Selectel


class Uploader:

    STORAGE_SELECTEL = 'selectel'

    def __init__(self, service: str, auth: dict = {}, options: dict = {}) -> None:
        if self.is_allowed_storage(service=service):
            self.service = service
        else:
            raise NotAllowed('Service ' + service + ' is not allowed')
        self.auth = auth
        self.options = options

    def is_allowed_storage(self, service: str) -> bool:
        return service in [self.STORAGE_SELECTEL]

    def upload(self, srcFilePath, dstName) -> str:
        if self.service == self.STORAGE_SELECTEL:
            selectel = Selectel(self.auth, self.options)
            return selectel.upload(srcFilePath=srcFilePath, dstFileName=dstName)



import time
from logging import Logger
from pbx_component_files_uploader.exceptions.errors import EmptyAuthField
from .token import Token

import requests


class Auth:
    AUTH_URL = 'https://auth.selcdn.ru/'

    def __init__(self, authData: dict, tokenCacheDir: str, requests: requests, logger: Logger):
        self.tokenCacheDir = tokenCacheDir
        if not tokenCacheDir:
            raise EmptyAuthField('token_cache_dir')
        self.check_auth_data(authData=authData)
        self.username = authData.get('username', '')
        self.password = authData.get('password', '')
        self.container = authData.get('container', '')

        self.requests = requests
        self.logger = logger
        self.token = Token(tokenCacheDir=tokenCacheDir, logger=logger).load()

    def check_auth_data(self, authData: dict) -> None:
        for k, v in authData.items():
            if not v:
                raise EmptyAuthField(k)

    def update_auth_token(self, new_token, storage_url, expires):
        self.token.secret = new_token
        self.token.storageUrl = storage_url
        self.token.expiresAt = int(time.time()) + int(expires)
        self.token.update()

    def authenticate(self):
        self.logger.debug('Need to authenticate with %s:%s',
                          self.username, self.password)
        resp = self.requests.get(self.AUTH_URL, headers={
            'X-Auth-User': self.username,
            'X-Auth-Key': self.password
        })
        if resp.status_code != 204:
            self.logger.debug(
                'Got an unexpected response from auth: %s', resp.content)
            raise Exception("Selectel: Unexpected status code: %s" %
                            resp.status_code)
        self.update_auth_token(
            resp.headers['X-Auth-Token'], resp.headers['X-Storage-Url'], resp.headers['X-Expire-Auth-Token'])

    def get_storage_data(self) -> dict:
        if self.token.is_expired():
            self.authenticate()

        return {
            'secret': self.token.secret,
            'storage_url': self.token.storageUrl,
            'container': self.container
        }



class Error(Exception):
    pass


class NotAllowed(Error):

    def __init__(self, message):
        self.message = message


class UploadError(Error):

    def __init__(self, code, message):
        self.code = code
        self.message = message


class EmptyAuthField(Error):

    def __init__(self, field):
        self.message = f'Auth field [{field}] is empty'


class EmptyConfigField(Error):

    def __init__(self, field):
        self.message = f'Config field [{field}] is empty'



from __future__ import annotations
from pbx_component_files_uploader.exceptions.errors import EmptyAuthField
from logging import Logger
from pathlib import Path
import time
import json


class Token:
    THRESHOLD = 300
    AUTH_TOKEN_STORAGE_FILE_NAME = 'token.selectel.json'

    def __init__(self, tokenCacheDir: str, logger: Logger):
        self.tokenCacheDir = tokenCacheDir
        if not tokenCacheDir:
            raise EmptyAuthField('token_cache_dir')
        self.logger = logger

    def load(self) -> Token:
        tokenFile = self.get_token_cache_file()

        tokenData = {'secret': '', 'storage_url': '',
                     'expires_at': int(time.time())}
        if tokenFile.is_file():
            try:
                tokenData = json.loads(tokenFile.read_text())
            except Exception as e:
                self.logger.debug('failed to load token cache from file')

        self.secret = tokenData['secret']
        self.storageUrl = tokenData['storage_url']
        self.expiresAt = tokenData['expires_at']

        return self

    def update(self) -> Token:
        self.get_token_cache_file().touch(exist_ok=True)
        self.get_token_cache_file().write_text(data=json.dumps({
            'secret': self.secret,
            'storage_url': self.storageUrl,
            'expires_at': self.expiresAt
        }), encoding='utf-8')

        return self

    def is_expired(self):
        return (self.expiresAt - int(time.time())) < self.THRESHOLD

    def get_token_cache_file(self) -> Path:
        return Path(self.tokenCacheDir + '/' + Token.AUTH_TOKEN_STORAGE_FILE_NAME)




class BaseUploader:

    def __init__(self, auth: dict, options: dict = {}) -> None:
        raise NotImplementedError

    def upload(self, srcFilePath: str, dstFileName: str) -> str:
        raise NotImplementedError
    """Upload file record for custom storage
    
    if file is uploaded, then return link for record
    if got errors while uploading raise UploadError exception
        
    """



from pathlib import Path
from multiprocessing import get_logger
from pbx_component_files_uploader.interfaces.base_uploader import BaseUploader
from pbx_component_files_uploader.exceptions.errors import UploadError
from .auth import Auth

import requests

log = get_logger()
log.name = 'selectel'


class Selectel(BaseUploader):

    DEFAULT_RETRIES = 3

    def __init__(self, authData: dict, options: dict = {}):
        self.maxRetries = options.get('max_retries', self.DEFAULT_RETRIES)
        self.auth = Auth(authData=authData, requests=requests,
                         tokenCacheDir=options.get('token_cache_dir', ''), logger=log)

    def build_url(self, storage_url, container, dstFile):
        return '{storage_url}/{name}/{key}'.format(
            storage_url=storage_url.rstrip('/'),
            name=container.strip('/'),
            key=dstFile.lstrip('/'),
        )

    def upload(self, srcFilePath: str, dstFileName: str) -> str:
        srcFile = Path(srcFilePath)
        if not srcFile.is_file():
            raise ValueError(f'No such file [{srcFilePath}]')

        if not dstFileName:
            raise ValueError('Empty [dstFileName] field')

        errorDesc = ''
        for i in range(self.DEFAULT_RETRIES):
            try:
                storageData = self.auth.get_storage_data()

                headers = {
                    'X-Auth-Token': storageData.get('secret')
                }
                url = self.build_url(storage_url=storageData.get(
                    'storage_url'), container=storageData.get('container'), dstFile=dstFileName)

                response = requests.put(
                    url=url, data=srcFile.open(mode='rb'), headers=headers)
                
                if response.status_code == 401:
                    self.auth.authenticate()

                if response.status_code == 201:
                    return response.url
            except Exception as e:
                errorDesc = str(e)
                log.debug(f'Attempt [{i + 1}]. Failed to upload. Cause [{errorDesc}]')

        raise UploadError(400, f'Failed upload file [{srcFilePath}]. Cause [{errorDesc}]')